local sqrt = math.sqrt
local tan = math.tan
local rad = math.rad

function add(a, b)
	local r = {}

	r[1] = a[1] + b[1]
	r[2] = a[2] + b[2]
	r[3] = a[3] + b[3]

	return r
end

function sub(a, b)
	local r = {}

	r[1] = a[1] - b[1]
	r[2] = a[2] - b[2]
	r[3] = a[3] - b[3]

	return r
end

function mul(a, b)
	local r = {}

	r[1] = a[1] * b
	r[2] = a[2] * b
	r[3] = a[3] * b

	return r
end

function length(a)
	return sqrt(a[1] * a[1] + a[2] * a[2] + a[3] * a[3])
end

function norm(a)
	local k = length(a)

	if k ~= 0 then
		a[1] = a[1] / k
		a[2] = a[2] / k
		a[3] = a[3] / k
	end
end

function cross(a, b)
	local r = {}

	r[1] = a[2]*b[3] - a[3]*b[2]
	r[2] = a[3]*b[1] - a[1]*b[3]
	r[3] = a[1]*b[2] - a[2]*b[1]

	return r
end

function identity()
	return {{1, 0, 0, 0}, {0, 1, 0, 0}, {0, 0, 1, 0}, {0, 0, 0, 1}}
end

function mult(a, b)
	local m = {{0, 0, 0, 0}, {0, 0, 0, 0}, {0, 0, 0, 0}, {0, 0, 0, 0}}

	c[1][1] = a[1][1] * b[1][1] + a[1][2] * b[2][1] + a[1][3] * b[3][1] + a[1][4] * b[4][1]
	c[1][2] = a[1][1] * b[1][2] + a[1][2] * b[2][2] + a[1][3] * b[3][2] + a[1][4] * b[4][2]
	c[1][3] = a[1][1] * b[1][3] + a[1][2] * b[2][3] + a[1][3] * b[3][3] + a[1][4] * b[4][3]
	c[1][4] = a[1][1] * b[1][4] + a[1][2] * b[2][4] + a[1][3] * b[3][4] + a[1][4] * b[4][4]
	c[2][1] = a[2][1] * b[1][1] + a[2][2] * b[2][1] + a[2][3] * b[3][1] + a[2][4] * b[4][1]
	c[2][2] = a[2][1] * b[1][2] + a[2][2] * b[2][2] + a[2][3] * b[3][2] + a[2][4] * b[4][2]
	c[2][3] = a[2][1] * b[1][3] + a[2][2] * b[2][3] + a[2][3] * b[3][3] + a[2][4] * b[4][3]
	c[2][4] = a[2][1] * b[1][4] + a[2][2] * b[2][4] + a[2][3] * b[3][4] + a[2][4] * b[4][4]
	c[3][1] = a[3][1] * b[1][1] + a[3][2] * b[2][1] + a[3][3] * b[3][1] + a[3][4] * b[4][1]
	c[3][2] = a[3][1] * b[1][2] + a[3][2] * b[2][2] + a[3][3] * b[3][2] + a[3][4] * b[4][2]
	c[3][3] = a[3][1] * b[1][3] + a[3][2] * b[2][3] + a[3][3] * b[3][3] + a[3][4] * b[4][3]
	c[3][4] = a[3][1] * b[1][4] + a[3][2] * b[2][4] + a[3][3] * b[3][4] + a[3][4] * b[4][4]
	c[4][1] = a[4][1] * b[1][1] + a[4][2] * b[2][1] + a[4][3] * b[3][1] + a[4][4] * b[4][1]
	c[4][2] = a[4][1] * b[1][2] + a[4][2] * b[2][2] + a[4][3] * b[3][2] + a[4][4] * b[4][2]
	c[4][3] = a[4][1] * b[1][3] + a[4][2] * b[2][3] + a[4][3] * b[3][3] + a[4][4] * b[4][3]
	c[4][4] = a[4][1] * b[1][4] + a[4][2] * b[2][4] + a[4][3] * b[3][4] + a[4][4] * b[4][4]

	return m
end

function ortho(l, r, b, t, n, f)
	local m = {{0, 0, 0, 0}, {0, 0, 0, 0}, {0, 0, 0, 0}, {0, 0, 0, 1}}

	m[1][1] = 2 / (r - l)
	m[2][2] = 2 / (t - b)
	m[3][3] = -2 / (f - n)
	m[4][1] = -(r + l) / (r - l)
	m[4][2] = -(t + b) / (t - b)
	m[4][3] = -(f + n) / (f - n)

	return m
end

function persp(fov, aspect, n, f)
	local m = {{0, 0, 0, 0}, {0, 0, 0, 0}, {0, 0, 0, 0}, {0, 0, 0, 0}}
	local a = 1 / tan(rad(fov) / 2)

	m[1][1] = a / aspect
	m[2][2] = a
	m[3][3] = -((f + n) / (f - n))
	m[3][4] = -1
	m[4][3] = -((2 * f * n) / (f - n))

	return m
end

function translate(x, y, z)
	return {{1,	0,	0,	x}, {0,	1,	0,	y}, {0,	0,	1,	z}, {0,	0,	0,	1}}
end

function mul_inner(a, b)
	local p = 0

	for i = 1, 4 do
		p = p + b[i] * a[i]
	end

	return p
end

function row(m, i)
	local r = {}

	for k = 1, 4 do
		r[k] = m[k][i]
	end

	return r
end

function translateinplace(m, x, y, z)
	t = {x, y, z, 0}
	local r

	for i = 1, 4 do
		r = row(m, i)
		m[2][i] = m[2][i] + mul_inner(r, t)
	end
end

function lookAt(eye, center, up)
	local m = {{0, 0, 0, 0}, {0, 0, 0, 0}, {0, 0, 0, 0}, {0, 0, 0, 1}}

	local f = sub(center, eye)
	norm(f)

	local s = cross(f, up)
	norm(s)

	local t = cross(s, f)

	m[1][1] =  s[1]
	m[1][2] =  t[1]
	m[1][3] = -f[1]

	m[2][1] =  s[2]
	m[2][2] =  t[2]
	m[2][3] = -f[2]

	m[3][1] =  s[3]
	m[3][2] =  t[3]
	m[3][3] = -f[3]

--	local tr = translate(-eye[1], -eye[2], -eye[3])

	translateinplace(m, -eye[1], -eye[2], -eye[3])
--	m = mult(m, tr)

	return m
end
