require "lin"

local lg = love.graphics
local rad = math.rad
local cos = math.cos
local sin = math.sin

local yaw		= -90
local pitch	= 0
local lastX
local lastY
local firstMouse = true

local pos = {0, 0, 3}
local up = {0, 1, 0}
local front = {0, 0, -1}

local camera
local screen
local mesh

local shader

local vs = [[extern mat4 model; extern mat4 projection; extern mat4 view;
		vec4 position(mat4 transform_projection, vec4 vertex_position)
		{
			return projection * view * model * vertex_position;
		}]]

local fs = [[vec4 effect(vec4 color, Image texture, vec2 texture_coords, vec2 screen_coords)
		{
				return texture2D(texture, texture_coords) * color;
		}]]

function love.load()
	width, height = lg.getDimensions()
	love.mouse.setPosition(width / 2, height / 2)
	lg.setBackgroundColor(20, 20, 20, 0)
	screen = lg.newCanvas(width, height)
	shader = lg.newShader(fs, vs)

	local image = lg.newImage("wall.jpg", {mipmaps = true})
	image:setWrap("repeat", "repeat")

	local num = 100
	local rnd = love.math.random
	local vertices =	{	{-num,	10, -num,	0,	0, rnd(), rnd(), rnd(), 1},
											{0,		10, -num,	num,	0, rnd(), rnd(), rnd(), 1},
											{0,		10, 0,	num,	num, rnd(), rnd(), rnd(), 1},
											{-num,	10, 0,	0,	num, rnd(), rnd(), rnd(), 1}
										}

	local format = {{"VertexPosition", "float", 3}, {"VertexTexCoord", "float", 2}, {"VertexColor", "byte", 4}}
	mesh = lg.newMesh(format, vertices, "triangles")

	local map = {1,2,3,1,3,4}
	mesh:setVertexMap(map)
	mesh:setTexture(image)
end

function update()
	lg.setCanvas(screen)
	lg.setShader(shader)

	shader:send("projection", "column", persp(45, width / height, 1.0, 1000.0))
	shader:send("view", "column", lookAt(pos, add(pos, front), up))

	lg.clear(0.2, 0.2, 0.2, 1)
	shader:send("model", "column", identity())
	lg.draw(mesh)

	lg.setShader()
	lg.setCanvas()
end

function love.draw()
	update()
	lg.draw(screen)
end

function love.update(dt)
	local	speed = 0.06

	if love.keyboard.isDown("w") then
		pos = add(pos, mul(front, speed))
	end
	if love.keyboard.isDown("s") then
		pos = sub(pos, mul(front, speed))
	end
	if love.keyboard.isDown("a") then
		tmp = cross(front, up)
		norm(tmp)
		tmp = mul(tmp, speed)
		pos = sub(pos, tmp)
	end
	if love.keyboard.isDown("d") then
		tmp = cross(front, up)
		norm(tmp)
		tmp = mul(tmp, speed)
		pos = add(pos, tmp)
	end
end

function love.mousemoved(x, y, dx, dy)
    if firstMouse then
			love.mouse.setPosition(width / 2, height / 2)
			love.mouse.setRelativeMode(true)
			lastX = x
			lastY = y
			firstMouse = false
		end

    xoffset = dx
    yoffset = dy
    lastX = x
    lastY = y

    sensitivity = 0.05
    xoffset = xoffset * sensitivity
    yoffset = yoffset * sensitivity

    yaw   = yaw + xoffset
    pitch = pitch + yoffset

    if pitch > 89 then
        pitch = 89
		end
    if pitch < -89 then
        pitch = -89
		end

    front[1] = cos(rad(yaw)) * cos(rad(pitch))
    front[2] = sin(rad(pitch))
    front[3] = sin(rad(yaw)) * cos(rad(pitch))
    norm(front)
		update()
end

function love.keypressed(key, scancode, isrepeat)
	if key == "escape" then
		love.event.quit()
	end
end
